# ToGA Artifact

This repository contains the replication artifact for TOGA: A Neural Method for Test Oracle Generation to appear in ICSE 2022.

Testing is widely recognized as an important stage of the softwaredevelopment lifecycle. Effective software testing can provide benefits such as documentation, bug finding, and preventing regressions. In particular, unit tests document a unit’s intended functionality. A test oracle, typically expressed as an condition, documents the intended behavior of the unit under a given test prefix. Synthesizing a functional test oracle is a challenging problem, as it has to capture the intended functionality and not the implemented functionality. In our paper, we propose TOGA (Test Oracle GenerAtion), a unified transformer-based neural approach to infer both exceptional and assertion test oracles based on the context of the focal method.

Our artifact reproduces the results for all RQs in the paper's evaluation.  The artifact includes source code and download links for datasets and models produced in the paper, fulfilling the requirements for reproduced, resuable, and available badges. We assume basic unix familiarity and ability to run python. Our artifact is given as a docker image for linux.

**Note:** For convenience, we provide a self-contained docker image to reproduce all results without any setup. We recommend using this to reproduce the results in the paper. See directions for using the docker image under `setup`.

## Organization

The core oracle generation tool is implemented in `toga.py` in the base directory. Other modules and scripts are organized in the following subdirectories:
- `model/`: modules and scripts for modeling. Used directly by `toga.py` for generating model inputs and running models to infer oracles.
  - `model/exception_data.py`: Utility methods for generating exception model inputs. The API call `exception_data.get_labeled_tests(tests, methods)` takes two lists of tests and their associated focal methods as input and returns four lists:`test_prefix, methods, labels, idxs`, which are respectively: tests stripped oracles, aligned focal methods, expected exception labels (0 or 1), and the index of labeled test triple into the original list (some tests may be dropped due to parsing errors).
  - `model/assertion_data.py`: Utility methods for generating assertion model inputs. The API call `assertion_data.get_model_inputs(tests, methods, vocab)` takes as input two aligned lists of tests and their associated focal methods, along with a vocab dictionary of common values for types that are used in generating candidate assertions. For example: `{int: [1, 10,...], float: [0.0, 1.1,...],...}`. This returns two lists with the format `[(focal_method, test_prefix, candidate_assert, label),...], idx`, where `focal_method`, `normalized_test`, and `candidate_assert` represent focal method, test prefix pairs associated with a possible assertion oracle. The final label value is unused during inference, but is kept for consistency with the training data format. The second list `idxs` represents the indexes of the prepared inputs, since some tests may be dropped due to parsing errors.
  - `model/exceptions/`: Scripts and code for running the exception model. Use scripts `run_train.sh` and `run_eval.sh` to train and run the model. 
  - `model/assertions/`: Scripts and code for running the assertion model. Use scripts `run_train.sh` and `run_eval.sh` to train and run the model.
- `scripts/`: scripts for generating datasets. See directions under `Datasets` section for usage to generate each dataset used in our evaluation.
- `eval/`: scripts for reproducing the evaluation from the TOGA paper. Note that rq3 uses the `toga.py` tool directly. See directions under `Evaluation` for usage to replicate each result in the paper.

`toga.py` usage: The toga tool takes (focal method, test) pairs as input and generates oracles using the two step inference procedure described in the paper. `toga.py` directly uses the modules and scripts under `model`.


## Setup
First, pull the docker image:    
`docker pull edinella/toga-artifact`

Alternately, if working with the downloading image tarball:
```
cat toga_docker.tar.gz.part-* > toga_docker.tar.gz
gzip -dk toga_docker.tar.gz && docker import toga_docker.tar edinella/toga-artifact
```

Connect to it: 
`docker run -i -t edinella/toga-artifact`

Then, setup some environment variables:    
```
export PATH=$PATH:/home/defects4j/framework/bin    
export ATLAS_PATH=/home/icse2022_artifact/data/atlas---deep-learning-assert-statements/
```

## Models    
The pretrained models are available in:
```
/home/icse2022_artifact/model/assertions/pretrained/    
/home/icse2022_artifact/model/exceptions/pretrained/    
```
Optionally, we provided scripts for training from scratch in `/home/icse2022_artifact/model/exceptions/run_train.sh` and `/home/icse2022_artifact/model/assertions/run_train.sh`. Training was performed on a vm with one P100 GPU.


### Defects4j setup (optional):
If you want to build and execute defects4j tests, [defects4j](https://github.com/rjust/defects4j) must be installed.

Requirements:
```
sudo apt install libdbi-perl
sudo apt install openjdk-8-jdk
sudo apt install libdbd-csv-perl
```

## Datasets - Preprocessed
Our approach is trained and evaluated on three datasets. The first, **Atlas\*** is an adaption of the [Atlas](https://gitlab.com/cawatson/atlas---deep-learning-assert-statements/) dataset. The preprocessed datasets **Atlas\*** and **Methods2Test\*** are included as `data/<DATASET>_star.tar.gz` files and be accessed by:
```
cd data
tar xzf atlas_star.tar.gz
tar xzf methods2test_star.tar.gz
```

The third dataset is our test dataset generated from Evosuite tests. This is located in `data/evosuite_tests.tar.gz`. Since the dataset size is very large (>400k tests), we provide two smaller sample datasets that can be used to reproduce the bug counts result and false positive rate result respectively from Table 3 for `Our Approach` in `data/evosuite_reaching_tests.tar.gz` and `data/evosuite_5project_tests.tar.gz`. `reaching_tests` contains bug-reaching tests only, while `5project_tests` contains the tests generated for the same 5 defects4j projects used in [1]'s evaluation.

To access the evosuite test datasets run:
```
cd data
tar xzf evosuite_reaching_tests.tar.gz
tar xzf evosuite_5project_tests.tar.gz
tar xzf evosuite_tests.tar.gz
```

## OPTIONAL Datasets - Generate from base datasets

```
cd data
python atlas_star_datagen.py
```

The resulting dataset will be stored in ``data/atlas_star``

Methods2Test\* Generation:
To generate **Methods2Test\***, first download the methods2test dataset).

```
git clone git@github.com:microsoft/methods2test.git
cd methods2test
export METHODS2TEST_PATH=`pwd`

cd dataset
for f in *.bz2; do tar xf $f; done
cd ../corpus/raw/fm
for f in *.bz2; do tar xf $f; done
```

Navigate to the methods2test_star_gen directory and execute scripts in order. Note that these scripts can take several hours to run because they must download all projects from the methods2test dataset and parse them for docstrings, so we recommend using the provided preprocessed dataset.

```
cd data/methods2test_star_generation
python gen_exception_dataset.py
bash download_projects.sh
python get_docstrings.py
python match_docstrings.py
```



## Evaluation
In our paper, we evaluate three research questions (RQs). 

1. **RQ1: Is our grammar representative of most developer-written assertions?**
   To evaluate this research question:    
   ``cd /home/icse2022_artifact/eval/rq1 && python rq1.py``

2. **RQ2: Can we infer assertions and exceptional behavior with high accuracy?**

**Exception Inference:**

To reproduce the exception results shown in table 1 for `TOGA Model`, run:
```
cd /home/icse2022_artifact/eval/rq2/exception_inference
bash rq2.sh
```

This script uses the pretrained exception model to predict whether a test is expected to trigger an exception or not, evaluated for accuracy and f1 score on the `methods2test_star` dataset. Note that the model used in the artifact has been retrained, so the results are slightly different from the submission (accuracy=85\% instead of 86\%, f1 score is 0.40 instead of 0.39). 

**Assertion Inference:**
To reproduce the exception results shown in table 2 for `TOGA Model`, run:
```
cd /home/icse2022_artifact/eval/rq2/assertion_inference
bash rq2.sh
```

This script uses the pretrained assertion model to predict an assertion given a test prefix and method under test's signature. we evaluate for accuracy and f1 score on the `atlas_star` dataset. 


To reproduce the weighted coin experiment in table 1, run:
```
cd eval/rq2/exception_inference
python coin.py
```


3. **RQ3: Can we catch bugs with low false alarms?**

To reproduce the results shown in Table 3 for `Our Approach`, run `toga.py` on either the bug-reaching inputs (for bug results) or 5 project sample (for false positive rate). By default, the `toga` tool will use the test metadata labels to evaluate oracles predicted by the models and print results. The tool also generates a `predicted_oracles.csv` file that can be used to generate executable test suites.

Note that we have improved our implementation since the submission and now find 3 additional bugs (57 total instead of 54).

To faciliate faster evaluation, the `toga.py` will automatically check its predicted oracles against labels included in its metadata input. This can save time since generating and executing test suites is potentially very time consuming. Note that `toga` can overestimate the `FP rate` when checking against labels, so the False Positive rate on generated tests will be lower.




To reproduce table 3 bug result by running only bug-reaching tests (this will not reproduce the FP rate, which requires running on all of the tests):
```
cd /home/icse2022_artifact/
python toga.py data/evosuite_reaching_tests/inputs.csv data/evosuite_reaching_tests/meta.csv
```

To reproduce table 3 false positive rate result on a 5 project sample (2+ hour runtime):
```
cd /home/icse2022_artifact/
python toga.py data/evosuite_5project_tests/inputs.csv data/evosuite_5project_tests/meta.csv
```


Both bug and FP rate results on the entire dataset (potentially 12+ hour runtime):
```
cd /home/icse2022_artifact/
python toga.py data/evosuite_5project_tests/inputs.csv data/evosuite_5project_tests/meta.csv
```



To validate the results, use the `/home/icse2022_artifact/eval/rq3/rq3.sh` to generate and run test suites from the toga generated oracles. By default, the script will generate tests for the bug-reaching tests:

```
cd /home/icse2022_artifact/eval/rq3/ && bash rq3.sh
```

The script can also be used to generate tests for a specified oracle prediction file, however on larger datasets this can have a very long runtime:
```
cd /home/icse2022_artifact/eval/rq3/ && bash rq3.sh /home/icse2022_artifact/data/evosuite_5project_tests/oracle_preds.csv
```



## References

1. Tufano, Michele, et al. "Unit Test Case Generation with Transformers." arXiv preprint arXiv:2009.05617 (2020). 
